/*
 * abissd/var.c - abissd, configuration variables
 *
 * Written 2004 by Werner Almesberger
 *
 * (C) Copyright Koninklijke Philips Electronics NV 2004. All rights reserved.
 * For licensing and warranty information, see the file ../../COPYING.LGPLv21
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>

#include <abissd.h>

#include "var.h"


struct var_user {
    const struct sched_ops *sched_ops;
    struct var_user *next;
};

static struct var {
    const char *name;
    const char *desc;
    struct var_user *users;
    char type;	/* 'i' for integer; 's' for string; 'f' for float (double) */
    int set;	/* 0 if unset; 1 if set */
    union {
	int i;
	double f;
	const char *s;
    } u;	/* undefined if !set */
    struct var *next;
} *vars = NULL;


/* ----- Lookups ----------------------------------------------------------- */


static struct var *var_find(const char *name)
{
    struct var *var;

    for (var = vars; var; var = var->next)
	if (!strcmp(name,var->name))
	    break;
    return var;
}


static struct var *var_require(const char *name)
{
    struct var *var;

    var = var_find(name);
    if (!var) {
	fprintf(stderr,"var_require: unknown variable \"%s\"\n",name);
	exit(1);
    }
    return var;
}


static struct var *var_require_type(const char *name,char type)
{
    struct var *var;

    var = var_require(name);
    if (var->type == type)
	return var;
    fprintf(stderr,"var_require_type: \"%s\" is of type \"%c\", want \"%c\"\n",
      name,var->type,type);
    exit(1);
}


/* ----- Registration and listings ----------------------------------------- */


static void add_user(struct var *var,const struct sched_ops *sched_ops)
{
    struct var_user *user;

    user = malloc(sizeof(struct var_user));
    user->sched_ops = sched_ops;
    user->next = var->users;
    var->users = user;
}


static void clear_users(struct var *var)
{
    while (var->users) {
	struct var_user *next;

	next = var->users->next;
	free(var->users);
	var->users = next;
    }
}


void var_register(const struct sched_ops *sched_ops,
  const char *name,char type,const char *desc)
{
    struct var *var;

    var = var_find(name);
    if (var) {
	if (var->type != type) {
	    fprintf(stderr,
	      "var_register: trying to change \"%s\" from \"%c\" to \"%c\"\n",
	      name,var->type,type);
	    exit(1);
	}
	if (var->users) {
	    if (sched_ops)
		add_user(var,sched_ops);
	    else
		clear_users(var);
	}
	return;
    }
    var = malloc(sizeof(*var));
    if (!var) {
	perror("malloc");
	exit(1);
    }
    var->name = strdup(name);
    var->desc = strdup(desc);
    var->users = NULL;
    if (sched_ops)
	add_user(var,sched_ops);
    if (!var->name || !var->desc) {
	perror("strdup");
	exit(1);
    }
    var->type = type;
    var->set = 0;
    var->next = vars;
    vars = var;
}


void var_show(FILE *file)
{
    struct var *var;

    for (var = vars; var; var = var->next) {
	if (!var->set)
	    continue;
	fprintf(file,"   %s = ",var->name);
	switch (var->type) {
	    case 'i':
		fprintf(file,"%d",var->u.i);
		break;
	    case 'f':
		fprintf(file,"%f",var->u.f);
		break;
	    case 's':
		fprintf(file,"\"%s\"",var->u.s);
		break;
	    default:
		abort();
	}
	fprintf(file,"  (%s)\n",var->desc);
    }
}


void var_usage(FILE *file)
{
    const struct var *var;
    const struct var_user *user;

    for (var = vars; var; var = var->next) {
	fprintf(file,"  %s=<%s>  %s",var->name,
	  var->type == 'i' ? "INT" :
	  var->type == 'f' ? "FLOAT" :
	  var->type == 's' ? "STR" : "???",
	  var->desc);
	if (var->users) {
	    for (user = var->users; user; user = user->next)
		fprintf(file,"%s%s",
		  user == var->users ? " (" : ", ",user->sched_ops->name);
	    fputc(')',file);
	}
	fputc('\n',file);
    }
}


/* ----- Assignment -------------------------------------------------------- */


void var_set_int(const char *name,int value)
{
    struct var *var;

    var = var_require_type(name,'i');
    var->u.i = value;
    var->set = 1;
}


void var_set_float(const char *name,double value)
{
    struct var *var;

    var = var_require_type(name,'f');
    var->u.f = value;
    var->set = 1;
}


void var_set_str(const char *name,const char *value)
{
    struct var *var;

    var = var_require_type(name,'i');
    if (var->set)
	free((char *) var->u.s);
    var->u.s = strdup(value);
    if (!var->u.s) {
	perror("strdup");
	exit(1);
    }
    var->set = 1;
}


void var_set(const char *name,const char *value)
{
    struct var *var;

    var = var_find(name);
    if (!var) {
	fprintf(stderr,"trying to set unknown variable \"%s\"\n",name);
	exit(1);
    }
    switch (var->type) {
	long tmp;
	char *end;

	case 'i':
	    tmp = strtol(value,&end,0);
	    if (*end || tmp < INT_MIN || tmp > INT_MAX) {
		fprintf(stderr,"invalid integer \"%s\"\n",value);
		exit(1);
	    }
	    var->u.i = tmp;
	    break;
	case 'f':
	    var->u.f = strtod(value,&end);
	    if (*end) {
		fprintf(stderr,"invalid float \"%s\"\n",value);
		exit(1);
	    }
	    break;
	case 's':
	    if (var->set)
		free((char *) var->u.s);
	    var->u.s = strdup(value);
	    if (!var->u.s) {
		perror("strdup");
		exit(1);
	    }
	    break;
	default:
	    abort();
    }
    var->set = 1;
}


void var_assignment(const char *expr)
{
    const char *eq;
    char *name;

    eq = strchr(expr,'=');
    if (!eq) {
	fprintf(stderr,"no \"=\" found in assignment\n");
	exit(1);
    }
    name = malloc(eq-expr+1);
    if (!name) {
	perror("malloc");
	exit(1);
    }
    strncpy(name,expr,eq-expr);
    name[eq-expr] = 0;
    var_set(name,eq+1);
    free(name);
}


/* ----- Retrieval --------------------------------------------------------- */


int var_is_set(const char *name)
{
    return var_require(name)->set;
}


int var_get_int(const char *name)
{
    struct var *var;

    var = var_require_type(name,'i');
    return var->u.i;
}


double var_get_float(const char *name)
{
    struct var *var;

    var = var_require_type(name,'f');
    return var->u.f;
}


const char *var_get_str(const char *name)
{
    struct var *var;

    var = var_require_type(name,'s');
    return var->u.s;
}
