/*
 * src/tools/ptree.c - Print a tree represented as a recursive list
 *
 * Written 2004 by Werner Almesberger
 *
 * (C) Copyright Koninklijke Philips Electronics NV 2004. All rights reserved.
 * For licensing and warranty information, see the file ../../COPYING.GPLv2
 */

/*
 * List syntax:
 *
 * - the list is enclosed by parentheses. The list may be embedded in other
 *   text, which is ignored. This text must not contain parentheses.
 *
 * - the list represents a tree structure. Each tree element is a list with
 *   the following structure: (<text>,<left>,<right>)
 *
 *   - <text> is text that may contain no commas. Parentheses are allowed.
 *
 *   - <left> and <right> are the left and right branches, respectively. They
 *     are in turn tree elements. If a branch is empty, there is no text for
 *     it. E.g. a leaf would look like this: (some text,,)
 */


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <sys/types.h>


#define READ_SIZE	1024
#define GAP		2	/* gap between adjacent items */


/*
 * Geometry:
 *
 *                        --------########--------------
 * <-----------indent------------>   |center          ...
 *                       |           |
 *                       |center    GAP
 *              +------------------+<->
 * <------indent------->###        |
 *              |       <--width-->|
 *              |    ##     ##     |
 *              :   ...     ...    :
 *              +------------------+
 *
 * "indent" is always measured from the left edge (and not from the adjacent
 * node).
 *
 * If the parent label is wider than the subtrees, things look as follows:
 *
 *         ####################################################
 * <indent>                          |center          ...
 *                       |           |
 *                       |center    GAP
 *              +------------------+<->
 * <------indent------->###        |
 *              |       <--width-->|
 *              +------------------+
 */


struct node {
    char *text;
    struct node *left;
    struct node *right;
    int indent;
    int width;
};


static char *input(void)
{
    char *buf = NULL;
    size_t size = 0;

    while (1) {
	ssize_t got;

	buf = realloc(buf,size+READ_SIZE);
	if (!buf) {
	    perror("realloc");
	    exit(1);
	}
	got = read(0,buf+size,READ_SIZE);
	if (got < 0) {
	    perror("read");
	    exit(1);
	}
	if (!got) {
	    buf[size] = 0;
	    return buf;
	}
	size += got;
    }
    /* not reached */
}


/*
 * No nice error handling - if anything goes wrong, we just crash on the spot.
 */

static struct node *treeify(char **s)
{
    char *end;
    struct node *node;

    assert(**s == '(');
    end = strchr(*s,',');
    *end = 0;
   
    node = malloc(sizeof(struct node));
    node->text = *s+1;

    end++;
    node->left = *end == ',' ? NULL : treeify(&end);
    end++;
    node->right = *end == ')' ? NULL : treeify(&end);
    *s = end+1;

    return node;
}


static void indent(struct node *node,int offset)
{
    int width_left,width_right;
    int center,offset_left,offset_right;

    /*
     * First, find the width of the trees underneath us.
     */
    if (node->left)
	indent(node->left,0);
    if (node->right)
	indent(node->right,0);
    width_left = node->left ? node->left->indent+node->left->width : 0;
    width_right = node->right ? node->right->indent+node->right->width : 0;

    center = strlen(node->text)/2-GAP/2;
    if (width_left >= center) {
	node->indent = offset+width_left-center;
	offset_left = offset;
    }
    else {
	node->indent = offset;
	offset_left = offset+center-width_left;
    }

    center = (strlen(node->text)+1)/2+(GAP+1)/2;
    offset_right = node->indent+center;
    if (width_right+center >= strlen(node->text))
	node->width = center+width_right;
    else
	node->width = strlen(node->text);

    /*
     * Now calculate the correct format.
     */
    if (node->left)
	indent(node->left,offset_left);
    if (node->right)
	indent(node->right,offset_right);
}


static int print_level(struct node *node,int depth,int goal,int *offset)
{
    int arm,i,end;

    if (depth < goal) {
	int more;

	more = node->left ? print_level(node->left,depth+1,goal,offset) : 0;
	more |= node->right ? print_level(node->right,depth+1,goal,offset) : 0;
	return more;
    }
#if 0
    printf("<%d,%d>",node->indent,node->width);
    fflush(stdout);
#endif
    arm = node->left ?
      node->left->indent+(strlen(node->left->text)/2) : node->indent;
    for (i = 0; i != node->indent-*offset; i++)
	putchar(i+*offset < arm ? ' ' : '-');

    printf("%s",node->text);

    end = node->indent+strlen(node->text);
    arm = node->right ?
      node->right->indent+(strlen(node->right->text)+1)/2 : end;
    if (arm < end)
	arm = end;
    for (i = end; i != arm; i++)
	putchar('-');
    *offset = arm;
    return 1;
}


static void print(struct node *root)
{
    int depth = 0;
    int offset = 0;

    while (print_level(root,0,depth,&offset)) {
	putchar('\n');
	offset = 0;
	depth++;
    }
}


int main(int argc,const char **argv)
{
    char *buf,*end;
    struct node *tree;

    switch (argc) {
	case 1:
	    buf = input();
	    break;
	case 2:
	    buf = strdup(argv[1]);
	    break;
	default:
	    fprintf(stderr,"usage: %s [list]\n",argv[0]);
	    return 1;
    }
    buf = strchr(buf,'(');
    end = strrchr(buf,')');
    end[1] = 0;
    tree = treeify(&buf);
    indent(tree,0);
    print(tree);
    return 0;
}
